/************************************************************************************/
/* 									Low Level MIDI Interface Driver								*/
/************************************************************************************/
/* Version: 1.0			Written by Chris Marshall September 22nd, 1989					*/
/* 										1989 by Chris Marshall										*/
/* Written using the THINK C version 4.0 development system.								*/
/************************************************************************************/
/*																												*/
/* This is a hardware level MIDI interface driver.  It provides direct access to 	*/
/* the SCC ports when connected to a MIDI interface that provides it's own clock.	*/
/* It is an interrupt-driven asynchronous driver that uses two circular buffers to	*/
/* queue the data being transferred.  It provides it's own routines to service the	*/
/* Level 2 secondary interrupts for the SCC port being used.  The driver MUST	be		*/
/* closed properly before it's memory is deallocated, as it replaces the vectors in	*/
/* the Lvl2DT table, and these must be replaced with the original ones.					*/
/* This driver doesn't do anything fancy, like time-stamp the MIDI data.  It's		*/
/* only reason for being is to get data from the outside world to your computer.		*/
/* What you do with it after that is your business.											*/
/* In order to do this, I have set up two circular buffers that are fed and emptied	*/
/* via two assembly language routines that are accessed through usurped LVL2DT		*/
/* vectors.  Reads and Writes to the driver interact with these buffers, as opposed	*/
/* to direct port access.																				*/
/*																												*/
/* If you really want to find out what's going on here, you should read that nasty	*/
/* little rag that Zilog supplies with their Z3580 chip.  It does tell you all that	*/
/* you need to know about the SCC, but you may have to read it over several times	*/
/* and consult Sister Zapposky's Tarot Emporium to figure it out.							*/
/*																												*/
/* This project took a lot of blood, sweat and tears, so I am offering it as			*/
/* example code, or even as a useable DRVR.  It works beautifully in Pluses, SEs		*/
/* and IIs.  I spent many late evenings reading that cheezy Zilog manual, and			*/
/* trying to keep from throwing a fit every time my Mac did.  I finally got it		*/
/* working, and working well, so enjoy.  If nothing else, it should be a good			*/
/* example of a DRVR.																					*/
/*																												*/
/* A note: This program uses a "hole" in the CODE segment to store a pointer to a	*/
/* private globals structure in a similar manner to THINK C's SetUpA4 () routine.	*/
/* The MyStash () routine returns this pointer.  I know it's pretty nasty, but it	*/
/* works like a charm!  Anyway, I've encapsuled it's access to one routine that can	*/
/* easily be modified to accomodate any better methods.										*/
/*																												*/
/* I also opted for a very illegal control call to adjust the size of the I/O			*/
/* buffers.  This can easily be changed to use the same method that the serial		*/
/* driver does.  All you would have to do is replace the buffer pointers with your	*/
/* own, and figure a way to properly dispose of the default buffers.  I suggest you	*/
/* keep auxilliary pointers to them, and dispose them at the driver's close.			*/
/*																												*/
/* This program is being distributed as a demo only!  If you plan to use it in a		*/
/* commercial product, please contact me at: AL:D5070, MARSHALL.CHR,						*/
/*	(301) 990-7773, or (301) 340-4573. - Chris Marshall										*/
/*																												*/
/* At the time of this writing, I am working for GE Information Services.				*/
/* 401 N. Washington St. M/S: MNC3J, Rockville, MD 20850										*/
/*																												*/
/* My home address is: 401 Blue Silk Lane Apt. C, Gaithersburg, Maryland 20879		*/
/*																												*/
/************************************************************************************/

#include "PortIO.h"				/* These are all of the constants used by the driver	*/
#include "MIDI.h"					/* Get the exported constants and declarations			*/
#include "Errors.h"				/* Heaven forbid we should need error codes!				*/

#define	DEFPORT		PORTA		/* The port this driver is for								*/
										/* PORTA = Modem port, PORTB = Printer port				*/

/************************************************************************************/
/* This is the main context for the MIDI driver.												*/
/************************************************************************************/

int main (pb, dce, n)
CntrlParam	*pb;					/*  Parameter block  											*/
DCtlEntry	*dce;					/*  Device control entry  										*/
int 			n;						/*  Entry point selector  										*/
	{
	switch (n)						/* Duh, which way do I go, George? Huh?					*/
		{
		case DRVROPEN:				/* Well, let's open da dwivah, Lenny!						*/
			return (OpenMyDriver (dce));
		break;
			
		case DRVRPRIME:
			return (MyDriverPrime ((HParamBlockRec *) pb));		/* Read or write 			*/
		break;
			
		case DRVRCONTROL:							/* Control call									*/
			return (MyControl ((CntrlParam *) pb));
		break;
			
		case DRVRSTATUS:							/* Status request									*/
			return (MyStatus ((CntrlParam *) pb));
		break;
			
		case DRVRCLOSE:							/* Close the driver								*/
			return (CloseMyDriver ());
		break;
		
		default:
			return (noErr);
		break;
		}
	}
   
/************************************************************************************/
/* This is the driver opening routine.  It allocates memory, and initializes all of */
/* the buffers and "private globals".  It will also replace the Lvl2DT vectors with */
/* interrupt handlers of it's own.																	*/
/************************************************************************************/

int OpenMyDriver (dce)
DCtlEntry		*dce;	/*  Device control entry  													*/
	{
	MyDataPtr	*stash;									/* Pointer to "private globals"		*/
	long			buffSize = DEFLEN, 					/* Buffer allocation size (default)	*/
					size;										/* Needed for allocation size			*/
	int			clock = DEFCLOCK;						/* Clock index (default)				*/
	int			err = noErr;							/* Error return 							*/
   
	/* Set up the driver flags																			*/
	dce -> dCtlFlags = (	dce -> dCtlFlags	/* Keep whatever is there						*/
								|	dReadEnable 	/* Driver is readable							*/
								| 	dWritEnable 	/* Driver is writeable							*/
								| 	dCtlEnable		/* Can do control calls							*/
								| 	dStatEnable		/* Can do status calls							*/
/*								|	dNeedGoodBye	/* Needs a goodbye smootch (no tounge)		*/
								| 	dNeedLock);		/* MUST be locked when in use					*/

	dce -> dCtlDelay = 0;			/* Zilch the delay											*/
	dce -> dCtlEMask = 0;			/* N/A for me													*/
	dce -> dCtlMenu = 0;				/* No menu either												*/
	dce -> dCtlWindow = NULL;		/* No window.													*/
						
   if (stash = MyStash ())			/* Get the "private globals" pointer					*/
   	{
   	if (!(*stash))					/* It will always be empty on the initial open		*/
   		{
   		/* First, allocate the main structure													*/
			if (*stash = (MyDataPtr) NewPtr ((long) sizeof (MyData)))
				{
				(*stash) -> inputBuffer = NULL;		/* Ensure that the buffer pointers	*/
				(*stash) -> outputBuffer = NULL;		/* are null, so we won't dispose		*/
				
				if (SetSize (DEFLEN) == noErr)		/* Successful allocations?				*/
					{
   				(*stash) -> outputBuffer -> buffSize = buffSize;	/* buffer size		*/
   					
  					(*stash) -> inputBuffer -> buffSize = buffSize;		/* buffer size		*/
   					
					AddressSCC ();						/* Set the addresses of the SCC			*/
					ClearBuffers ();					/* Initialize the buffers					*/
					
					err = noErr;						/* All clear									*/
					}
				else
					err = memFullErr;					/* Return the system memory full error	*/
				}
			else
				err = memFullErr;		/* Return the system memory full error					*/
			}
		else				/* Driver is already open													*/
			{
			err = noErr;				/* Return no error											*/
			}
		}
	else
		err = STASHERR;			/* Couldn't find the stash (should NEVER happen)		*/
			
	if (err != noErr)
		err = OPENBAD;				/* Return a generic open error for any opening errors	*/
		
	return (err);
	}
   
/************************************************************************************/
/* This is the driver's Prime routine.  It handles read and write operations on the */
/* buffers that are filled and emptied on an interrupt basis via the GetOne () and  */
/* PutOne () operations.                                                            */
/************************************************************************************/

int MyDriverPrime (pb)
HParamBlockRec		*pb;
	{
	int	err = noErr;				/* Error return												*/
	
	switch ((unsigned char) (pb -> ioParam.ioTrap & 0x00FF))	/* What should I do?		*/
		{
		case aRdCmd:					/* This is a read operation								*/
			if (ReadData (pb))		/* Look out for read errors.								*/
				err = READBAD;
		break;
			
		case aWrCmd:					/* This is a write operation								*/
			if (WriteData (pb))		/* Look out for write errors.								*/
				err = WRITEBAD;
		break;
			
		default:							/* Oops!  We can't make sense of this.					*/
			err = PRIMEBAD;
		break;
		}
	
	return (err);
	}
   
/************************************************************************************/
/* This is the driver's control routine.  It can be used to change the external		*/
/* clock rate index, and the primary buffer sizes.  It can also be used to "flush"	*/
/* the primary buffers.																					*/
/************************************************************************************/

int MyControl (pb)
CntrlParam	*pb;
	{
	int			err = noErr;
	long			size;
	MyDataPtr	*stash;
	
	stash = MyStash ();					/* Get our private data									*/
	
	switch (pb -> csCode)
		{
		case	RESETALL:	/* Reset the registers to default, and flush the buffers		*/
		case	KILLIO:		/* Same thing to stop I/O												*/
		case	FLUSHBUFF:	/* Flush the buffers														*/
			(*stash) -> driverErr = ClearBuffers ();		/* Clear or indicate error		*/

		break;
			
		case	RESIZE:		/* Change the size of the primary buffers							*/
								/* csParam contains the new size										*/
								/* The size is a long integer that									*/
								/* is cast between the first two										*/
								/* array elements.														*/
								/* WARNING: This routine uses memory manager calls.  Async	*/
								/* is a no-no!																*/
								
			size = *((long *) pb -> csParam);
			
			(*stash) -> driverErr = SetSize (size);	/* Clear or indicate error			*/

		break;
			
		case	CLOCKRATE:	/* Change the clock source index.									*/
   		DoRegisters (pb -> csParam [0]);			/* csParam contains the new index	*/
																/* 0 = 31.25 KHz clock source			*/
																/* 1 = 500 KHz clock source			*/
																/* 2 = 1 MHz clock source				*/
																/* 3 = 2 MHz clock source				*/

			(*stash) -> driverErr = err;				/* Clear or indicate error				*/
		break;
		
		case	GOODBYE:										/* Need to replace the vectors!		*/
			RestoreRegisters ();							/* Restore the SCC registers			*/
			RestoreVectors ();							/* Restore the Lvl2DT vectors			*/
		break;
		}
	
	return (err);
	}

/************************************************************************************/
/* This is the driver's Status routine.  It will return the external clock rate		*/
/* index, the primary buffer sizes, and status of each buffer.  It will also return	*/
/* the contents of the port's SCC read registers.												*/  
/************************************************************************************/

int MyStatus (pb)
CntrlParam *pb;
	{
	MyDataPtr	*stash;												/* Our private data			*/
	MyParms		*statReturn = (MyParms *) pb -> csParam;	/* We'll cast it				*/
	
	switch (pb -> csCode)								/* What do they want?					*/
		{
		case CHECKREG:										/* SCC Register check?					*/
			pb -> csParam [0] = (int) ReadCtlReg (0);	/* Read each register				*/
			pb -> csParam [1] = (int) ReadCtlReg (1);
			pb -> csParam [2] = (int) ReadCtlReg (2);
			pb -> csParam [3] = (int) ReadCtlReg (3);
			pb -> csParam [4] = (int) ReadCtlReg (8);
			pb -> csParam [5] = (int) ReadCtlReg (10);
			pb -> csParam [6] = (int) ReadCtlReg (12);
			pb -> csParam [7] = (int) ReadCtlReg (13);
			pb -> csParam [8] = (int) ReadCtlReg (15);
			pb -> csParam [9] = DEFPORT;				/* Return the driver's port.	1 = A	*/
																/*										0 = B	*/
		break;
		
		case STATREC:
		default:
			if (stash = MyStash ())		/* Get a pointer to our "private stash"			*/
				{
				statReturn -> myClock = (*stash) -> myClock;
				statReturn -> inBufferSize = (*stash) -> inputBuffer -> buffSize;
				statReturn -> outBufferSize = (*stash) -> outputBuffer -> buffSize;
				statReturn -> inBufferChar = (*stash) -> inputBuffer -> count;
				statReturn -> outBufferChar = (*stash) -> outputBuffer -> count;
				statReturn -> driverErr = (*stash) -> driverErr;
				}
			else
				{
				return (STASHERR);
				}
		break;
		}
		
	return (noErr);
	}

/************************************************************************************/
/* This is the driver closing routine.  It deallocates all of the buffers and the	*/
/* private data, and restores the original Lvl2DT vectors.									*/
/************************************************************************************/

int CloseMyDriver ()
	{
	MyDataPtr	*stash;
	
	if (stash = MyStash ())
		{
		ResetRegisters ();
		
		if ((*stash) -> inputBuffer)
			{
			DisposPtr ((*stash) -> inputBuffer);
			}
			
		if ((*stash) -> outputBuffer)
			{
			DisposPtr ((*stash) -> outputBuffer);
			}

		DisposPtr (*stash);
		
		*stash = 0L;
		}

	return (noErr);
	}

/************************************************************************************/
/* This routine will set the size of the buffers to that passed in.						*/
/************************************************************************************/

int SetSize (size)
long	size;
	{
	MyDataPtr	*stash;
	long			origSize = size;
	
	size = (long) (size + sizeof (MyBuffer) + (size % 2));
   			
	if (stash = MyStash ())			/* If we have a stash...									*/
		{
		if (*stash)						/* If there is already something there...				*/
			{
			if ((*stash) -> inputBuffer)	/* Is there an input buffer allocated?			*/
				{
				DisposPtr ((*stash) -> inputBuffer);	/* Get rid of it						*/
				}
				
			if ((*stash) -> outputBuffer)		/* Is there an output buffer allocated?	*/
				{
				DisposPtr ((*stash) -> outputBuffer);	/* Get rid of it						*/
				}
			}
		
   	/* Allocate the input buffer																	*/
		if ((*stash) -> inputBuffer = (MyBuffer *) NewPtr (size))
			{
			/* Allocate the output buffer																*/
			if (((*stash) -> outputBuffer = (MyBuffer *) NewPtr (size)))
				{
				(*stash) -> inputBuffer -> buffSize = 
				(*stash) -> outputBuffer -> buffSize = origSize;
				
				(*stash) -> inputBuffer -> count = (*stash) -> outputBuffer -> count = 0L;
				
				(*stash) -> inputBuffer -> fill = (*stash) -> inputBuffer -> buff;
				(*stash) -> inputBuffer -> empty = (*stash) -> inputBuffer -> buff;
				
				(*stash) -> outputBuffer -> fill = (*stash) -> outputBuffer -> buff;
				(*stash) -> outputBuffer -> empty = (*stash) -> outputBuffer -> buff;
				
				return (noErr);
				}
			else
				{
				DisposPtr ((*stash) -> inputBuffer);	/* Get rid of the previous alloc	*/
				return (memFullErr);	/* Return the system memory full error					*/
				}
			}
		else
			{
			return (memFullErr);		/* Return the system memory full error					*/
			}
		}
	else
		return (STASHERR);			/* This would be a serious problem						*/
	}

/************************************************************************************/
/*	This routine will reset the buffers' fill and empty pointers, and their counts.	*/   
/************************************************************************************/

int ClearBuffers ()
	{
	MyDataPtr	*stash;	/* Pointer to our private data										*/
	
	if (stash = MyStash ())			/* If we have a stash...									*/
		{
		if (*stash)						/* And it points to something...							*/
			{
 			(*stash) -> driverErr = noErr;	/* No errors										*/

			(*stash) -> outputBuffer -> fill = &((*stash) -> outputBuffer ->buff [0]);
   		/* Set fill pointer to start of buffer													*/
   		
			(*stash) -> outputBuffer -> empty = &((*stash) -> outputBuffer ->buff [0]);
   		/* Set empty pointer to start of buffer												*/
   		
			(*stash) -> outputBuffer -> count = 0L;
   		/* Initialize the byte count to 0														*/
   		
			(*stash) -> inputBuffer -> fill = &((*stash) -> inputBuffer -> buff [0]);	
   		/* Set fill pointer to start of buffer													*/
   						
			(*stash) -> inputBuffer -> empty = &((*stash) -> inputBuffer -> buff [0]);
   		/* Set empty pointer to start of buffer												*/
   						
			(*stash) -> inputBuffer -> count = 0L;
   		/* Initialize the byte count to 0														*/
			
			SetUpRegisters ();				/* Set SCC registers up						*/
			}
		else
			return (NOSTASH);			/* There is no allocated stash!							*/
		}
	else
		return (STASHERR);			/* This would be a serious problem						*/

	return (noErr);					/* Like s--- through a goose								*/
	}

/************************************************************************************/
/* This is the read operation for the Prime routine.											*/
/************************************************************************************/

int ReadData (pb)
HParamBlockRec	*pb;
	{
	Ptr			buffer = pb -> ioParam.ioBuffer;		/* The buffer to be filled			*/
	long			byteReq = pb -> ioParam.ioReqCount;	/* How much they want				*/
	int			err = noErr;								/* Error return						*/
	MyDataPtr	*stash;										/* Pointer to private data			*/
	
	pb -> ioParam.ioActCount = ReadBuff ((unsigned char *) pb -> ioParam.ioBuffer, 
													pb -> ioParam.ioReqCount);
	/* Read from the primary buffer.																	*/
	
	if (stash = MyStash ())			/* Get the private data pointer							*/
		{
		if ((*stash) -> driverErr == INBUFFOV)			/* Check for possible data stomp	*/
			err = INBUFFOV;
		}
		
	return (err);
	}

/************************************************************************************/
/* This is the data writing routine for the Prime routine.									*/
/************************************************************************************/

int WriteData (pb)
HParamBlockRec	*pb;
	{
	pb -> ioParam.ioActCount = WriteBuff ((unsigned char *) pb -> ioParam.ioBuffer, 
													pb -> ioParam.ioReqCount);
	
	if (pb -> ioParam.ioActCount != pb -> ioParam.ioReqCount)
		{
		return (OUTBUFFOV);
		}
	
	return (noErr);
	}

/************************************************************************************/
/* This routine reads req ammount of bytes from the primary input buffer.  If there	*/
/* are not enough bytes, no error is generated.  The number of bytes actually read  */
/* is returned in the function return.																*/
/************************************************************************************/

long	ReadBuff (buff, req)
unsigned char	*buff;						/* Pointer to receiving buffer					*/
long				req;							/* How many bytes they want						*/
	{
	MyDataPtr		*stash;					/* Access the primary buffer						*/
	long				ret = 0L;				/* Number of bytes actually read					*/
	unsigned char	*end;						/* Will point to the end of the primary buff	*/
	
	if (!buff || !req)						/* If there is no buffer, or no request...	*/
		return (0L);							/* ...Get outta Dodge								*/
		
	if (stash = MyStash ())					/* Get private data									*/
		{
		end = ((*stash) -> inputBuffer -> buff 				/* Point to end of buffer	*/
				+ (*stash) -> inputBuffer -> buffSize) - 1L;
		
		/******************************************************************************/
		/* This may look incredibly redundant, but remember we are dealing with			*/
		/* circular buffers that are being fed via interrupt.  The primary buffer		*/
		/* could be getting fed as quickly as it's being emptied!  We have to make		*/
		/* sure that our data is as current as possible.  This does not absoluteley	*/
		/* guarantee integrity; but it's pretty good.											*/
		/******************************************************************************/
		
		while ((*stash) -> inputBuffer -> count && req)	
		/* As long as we got both bytes and room...												*/
			{
			*buff++ = *((*stash) -> inputBuffer -> empty)++;	/* Transfer the byte		*/
			
			if ((*stash) -> inputBuffer -> empty > end)	/* See if we're at the end		*/
				(*stash) -> inputBuffer -> empty = 	/* Go back to beginning of buffer	*/
					(*stash) -> inputBuffer -> buff;
					
			ret++;								/* Increment the count								*/
			(*stash) -> inputBuffer -> count--;	/* Decrement the byte count				*/
			req--;										/* Decrement the requested bytes			*/
			}
		}
	
	return (ret);
	}

/************************************************************************************/
/* This is the Prime call's writing routine.  It fills the primary output buffer.	*/
/************************************************************************************/

long WriteBuff (buff, count)
unsigned char	*buff;
long				count;
	{
	MyDataPtr		*stash;					/* Access the primary buffer						*/
	unsigned char	*ptr;						/* This will point to the end of the buffer	*/
	long				ret = 0L;				/* The actual count of bytes written			*/

   if (stash = MyStash ())				/* Get the private data									*/
   	{
   	ptr = ((*stash) -> outputBuffer -> buff + (*stash) -> outputBuffer -> buffSize);
   	while (count--)					/* While we still have data...						*/
   		{
   		(*stash) -> outputBuffer -> count++;	/* Increment the byte count			*/
   		*(*stash) -> outputBuffer -> fill++ = *buff++;	/* Transfer the byte			*/
   		if ((*stash) -> outputBuffer -> fill > ptr)		/* At the end?					*/
   			(*stash) -> outputBuffer -> fill = (*stash) -> outputBuffer -> buff;
   			/* Reset the pointer to the beginning												*/
   			
   		/* Are we asking too much?																	*/
   		if ((*stash) -> outputBuffer -> count > (*stash) -> outputBuffer -> buffSize)
   			break;												/* Return early					*/
   			
   		ret++;
   		}
   	
   	PutOne ();							/* Kick-start the sucker								*/
   	}
   
   return (ret);
	}

/************************************************************************************/
/* This routine replaces the secondary interrupt vectors with pointers to the input	*/
/* and output routines, and saves the old vectors in our "private" space.				*/
/************************************************************************************/

int ReplaceVectors ()
   {
   MyDataPtr	*stash;

   if (stash = MyStash ())	/* Get the location of our "private stash"					*/
   	{
		WriteCtlReg (9, 0);						/* Disable any spurious interrupts			*/
   		
   	(*stash) -> oldVector1 = (Ptr) (Lvl2DT [TBEOFFSET]);
   	(*stash) -> oldVector2 = (Ptr) (Lvl2DT [CAVOFFSET]);		/* Save old vectors	*/
   	
   	Lvl2DT [TBEOFFSET] = (ProcPtr) PutOne;			/* Place the output vector			*/
   	Lvl2DT [CAVOFFSET] = (ProcPtr) GetOne;			/* Place the input vector			*/
   	}
   else							/* Omigoshgolli														*/
   	{
   	return (STASHERR);	/* This should NEVER happen										*/
   	}
   
   return (noErr);
   }

/************************************************************************************/
/* This routine will put the old vectors back into the vector table.						*/
/************************************************************************************/

int RestoreVectors ()
   {
   MyDataPtr	*stash;
   
   if (stash = MyStash ())	/* Get the location of our "private stash"					*/
   	{
		WriteCtlReg (9, 0);						/* Disable any spurious interrupts			*/
   	
   	Lvl2DT [TBEOFFSET] = (ProcPtr) (*stash) -> oldVector1;	/* Restore the 		*/
   	Lvl2DT [CAVOFFSET] = (ProcPtr) (*stash) -> oldVector2;	/* old vectors			*/
   	}
   else							/* Omigoshgolli														*/
   	{
   	return (STASHERR);	/* This should NEVER happen										*/
   	}
   
   return (noErr);
   }
   
/************************************************************************************/
/* This routine sets the registers of the SCC to the form needed by the MIDI driver.*/
/************************************************************************************/

int SetUpRegisters ()
	{
	DoRegisters (DEFCLOCK);				/* Set up the channel SCC registers					*/
	ReplaceVectors ();					/* Replace the Lvl2DT vectors							*/

   return (noErr);
   }
   
/************************************************************************************/
/* This routine will reset the SCC to a condition useable by the serial drivers.		*/
/************************************************************************************/

int ResetRegisters ()
	{
	RestoreRegisters ();					/* Reset the SCC											*/
	RestoreVectors ();					/* Put the old Lvl2DT vectors back					*/

   return (noErr);
	}

/************************************************************************************/
/* This routine sets the various values of the SCC addresses into the stash.			*/
/************************************************************************************/

int AddressSCC ()
	{
	MyDataPtr 		*stash;	/* Points to our private stash									*/
   
   if (stash = MyStash ())	/* Get the stash pointer											*/
   	{
		(*stash) -> sccDataRd = ((unsigned char *) SCCRd) + DATOF;
		(*stash) -> sccCtrlRd = ((unsigned char *) SCCRd) + CTLOF;
		(*stash) -> sccDataWr = ((unsigned char *) SCCWr) + DATOF;
		(*stash) -> sccCtrlWr = ((unsigned char *) SCCWr) + CTLOF;
		
		return (noErr);
   	}
   else
   	{
   	return (STASHERR);
   	}
   	
	return (noErr);
	}
   
   
/************************************************************************************/
/* This is a blanket register setup routine.  Pass in the clock index.					*/
/************************************************************************************/

int DoRegisters (clock)
int	clock;							/* The clock index											*/
	{
	MyDataPtr		*stash;										/* Our private data				*/
	unsigned char	reg4 = REG4;								/* Specify 1 stop bit 			*/
   unsigned char	reset = 1;									/* This will be a chan reset	*/

	if (!(stash = MyStash ()))
		return (STASHERR);
		
	if (!*stash)
		return (NOSTASH);
	
	(*stash) -> myClock = clock;				/* Record the index								*/

	reset = (reset << (6 + DEFPORT));		/* Create our channel reset					*/

/* First: Reset the channel */

	WriteCtlReg (9, reset);						/* Reset the channel								*/

/* Next: set the conditions */

	switch (clock)			/* Determine the divide-by clock mode 								*/
      {
		case CL31KHZ:   
		break;           	/* 31.25 KHz clock frequency     									*/
         
		case CL500KHZ:
			reg4 |= 0x40;	/* 500 KHz clock                 									*/ 
		break;
      
		case CL1MHZ:
			reg4 |= 0x80;	/* 1 MHz clock                   									*/
		break;
         
		case CL2MHZ:
			reg4 |= 0xC0;	/* 2MHz clock                    									*/
		break;
      }

	WriteCtlReg (4, reg4);				/* Set the clock											*/
	WriteCtlReg (1, REG1);				/* Disable our interrupts for Rx, Tx				*/
	WriteCtlReg (2, REG2);				/* No vector												*/
	WriteCtlReg (3, REG3);				/* Rx 8 bits/char.										*/
	WriteCtlReg (5, REG5);				/* Tx 8 bits/char RTS enabled							*/
	WriteCtlReg (6, REG6);				/* No sync stuff											*/
	WriteCtlReg (7, REG7);				/* Ditto														*/
	WriteCtlReg (10, REG10);			/* Nuthin' fancy											*/
	WriteCtlReg (11, REG11);			/* Clock sources are the TRxC pins					*/
	WriteCtlReg (12, REG12);			/* No Time Constant										*/
	WriteCtlReg (13, REG13);			/* Ditto														*/
	WriteCtlReg (14, REG14);			/* Disable the baud rate generator					*/

/* Next: Set the enables */

	WriteCtlReg (3, REG3B);				/* Rx															*/
	WriteCtlReg (5, REG5B);				/* Tx															*/

/* Finally: The interrupts */

	WriteCtlReg (15, REG15);			/* Reset the DCD interrupt								*/
	WriteCtlReg (0, RESETTBE);			/* Reset the transmit buffer MT interrupt			*/
	WriteCtlReg (0, RESETEXT);			/* Reset the EXT interrupt								*/
	WriteCtlReg (0, RESETEXT);			/* Reset the EXT interrupt								*/
	WriteCtlReg (1, REG1B);				/* Set up for Rx, Tx and EXT interrupts			*/
	WriteCtlReg (9, REG9);				/* "Throw the switch, Igor!" - "Yes, master!"	*/
	
	return (noErr);
	}
   
/************************************************************************************/
/* This routine sets the SCC back in a semi-useable state.									*/
/************************************************************************************/

int RestoreRegisters ()
	{
   unsigned char	reset = 1;									/* This will be a chan reset	*/

	reset = (reset << (6 + DEFPORT));		/* Create our channel reset					*/

/* First: Reset the channel */

	WriteCtlReg (9, reset);						/* Reset the channel								*/

/* Next: The interrupts */

	WriteCtlReg (15, REG15);			/* Reset the DCD interrupt								*/
	WriteCtlReg (0, RESETEXT);			/* Reset the EXT interrupt								*/
	WriteCtlReg (0, RESETEXT);			/* Reset the EXT interrupt								*/
	WriteCtlReg (1, EXTONLY);			/* Set up for EXT interrupts only					*/
	WriteCtlReg (9, REG9);				/* "Throw the switch, Igor!" - "Yes, master!"	*/
	
	return (noErr);
	}
   
/************************************************************************************/
